/* bconf.c - Configuration interpreter for the new bind style conf.
**
** Some code derived from csircd and Sentinel
**
** $Id: bconf.c,v 1.37 2003/11/05 01:49:23 wcampbel Exp $
*/

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "config.h"
#include "cparse.h"
#include "bconf.h"
#include "main.h"
#include "me.h"
#include "mem.h"
#include "servers.h"
#include "send.h"
#include "privs.h"
#include "dlink.h"
#include "operserv.h"
#include "net.h"
#include "irc_string.h"
#include "protected.h"
#include "services.h"

struct ConfTable
{
  char *name;
  int rehashable;
  int (*handler)(CONFIGENTRY *);
};

static int generic_subblock_handler(CONFIGENTRY *, char *, struct ConfTable *);

static int c_connect(CONFIGENTRY *);
static int c_oper(CONFIGENTRY *);
static int c_serverinfo(CONFIGENTRY *);
static int c_settings(CONFIGENTRY *);
static int c_channels(CONFIGENTRY *);
static int c_operserv(CONFIGENTRY *);
static int c_reservednick(CONFIGENTRY *);
static int c_protected(CONFIGENTRY *);
static int c_services(CONFIGENTRY *);
static struct ConfTable conf_root_table[] =
  {
    { "CONNECT", 1, c_connect },
    { "OPER", 1, c_oper },
    { "OPERATOR", 1, c_oper },
    { "SERVERINFO", 1, c_serverinfo },
    { "SETTINGS", 1, c_settings },
    { "CHANNELS", 0, c_channels },
    { "OPERSERV", 0, c_operserv },
    { "RESERVEDNICK", 1, c_reservednick },
    { "PROTECTED", 1, c_protected },
    { "SERVICES", 1, c_services },
    { NULL, 0, NULL }
  };

static int c_si_name(CONFIGENTRY *);
static int c_si_ip(CONFIGENTRY *);
static int c_si_description(CONFIGENTRY *);
static struct ConfTable conf_serverinfo_table[] =
  {
    { "NAME", 0, c_si_name },
    { "IP", 1, c_si_ip },
    { "DESCRIPTION", 0, c_si_description },
    { NULL, 0, NULL }
  };

static int c_os_nick(CONFIGENTRY *);
static int c_os_username(CONFIGENTRY *);
static int c_os_gecos(CONFIGENTRY *);
static struct ConfTable conf_operserv_table[] =
  {
    { "NICK", 0, c_os_nick },
    { "USERNAME", 0, c_os_username },
    { "GECOS", 0, c_os_gecos },
    { NULL, 0, NULL }
  };

static int c_se_opernotice(CONFIGENTRY *);
static int c_se_ircdtype(CONFIGENTRY *);
static int c_se_operoperwall(CONFIGENTRY *);
static int c_se_cloneurl(CONFIGENTRY *);
static int c_se_nickserv(CONFIGENTRY *);
static int c_se_nickpass(CONFIGENTRY *);
static int c_se_floodrate(CONFIGENTRY *);
static int c_se_nicklen(CONFIGENTRY *);
static int c_se_conntimeout(CONFIGENTRY *);
static int c_se_authtimeout(CONFIGENTRY *);
static int c_se_use_settime(CONFIGENTRY *);
static int c_se_clone_limit(CONFIGENTRY *);
static int c_se_clone_gline(CONFIGENTRY *);
static int c_se_clone_gline_time(CONFIGENTRY *);
static int c_se_clone_over(CONFIGENTRY *);
static int c_se_fclone_gline_time(CONFIGENTRY *);
static int c_se_maxkill(CONFIGENTRY *);
static int c_se_default_echo(CONFIGENTRY *);
static struct ConfTable conf_settings_table[] =
  {
    { "OPERNOTICE", 1, c_se_opernotice },
    { "IRCDTYPE", 0, c_se_ircdtype },
    { "OPEROPERWALL", 1, c_se_operoperwall },
    { "CLONEURL", 1, c_se_cloneurl },
    { "NICKPASS", 0, c_se_nickpass },
    { "NICKSERV", 0, c_se_nickserv },
    { "FLOODRATE", 1, c_se_floodrate },
    { "NICKLEN", 1, c_se_nicklen },
    { "CONNTIMEOUT", 1, c_se_conntimeout },
    { "AUTHTIMEOUT", 1, c_se_authtimeout },
    { "USE_SETTIME", 1, c_se_use_settime },
    { "CLONE_LIMIT", 1, c_se_clone_limit },
    { "CLONE_GLINE", 1, c_se_clone_gline },
    { "CLONE_GLINE_TIME", 1, c_se_clone_gline_time },
    { "CLONE_OVER", 1, c_se_clone_over },
    { "CLONE_FCLONE_GLINE_TIME", 1, c_se_fclone_gline_time },
    { "DEFAULT_ECHO", 1, c_se_default_echo },
    { "MAXKILL", 1, c_se_maxkill },
    { NULL, 0, NULL }
  };

void ConfLoad(void)
{
  CONFIGFILE *cfptr;
  CONFIGFILE *cfp;
  CONFIGENTRY *ce;
  struct ConfTable *ct = NULL;

  cfptr = cfp = config_load(CONF);
  if (cfp == NULL)
  {
    fprintf(stderr, "Unable to open configuration file (%s): %s\n",
            CONF, strerror(errno));
    if (me.logfd != NULL)
      fclose(me.logfd);
    exit(0);
  }
  for (;cfptr;cfptr = cfptr->cf_next) /* Handle include files? */
  {
    for (ce = cfptr->cf_entries;ce;ce = ce->ce_next)
    {
      for (ct = conf_root_table; ct->name; ct++)
      {
        if (!strcasecmp(ct->name, ce->ce_varname))
        {
          ct->handler(ce);
          break;
        }
      }
    }
    if (ct->name == NULL)
    {
      slog("%s:%i: invalid config option: %s",
           ce->ce_fileptr->cf_filename, ce->ce_varlinenum,
           ce->ce_varname);
    }
  }
  config_free(cfp);
}

static int generic_subblock_handler(CONFIGENTRY *ce, char *subblock,
                                    struct ConfTable *table)
{
  struct ConfTable *ct = NULL;

  for (ce = ce->ce_entries;ce;ce = ce->ce_next)
  {
    for (ct = table; ct->name; ct++)
    {
      if (!strcasecmp(ct->name, ce->ce_varname))
      {
        ct->handler(ce);
        break;
      }
    }
    if (ct->name == NULL)
    {
      slog("%s:%i: invalid %s config option: %s",
           ce->ce_fileptr->cf_filename, ce->ce_varlinenum,
           subblock, ce->ce_varname);
    }
  }
  return 0;
}

static int c_connect(CONFIGENTRY *ce)
{
  char *name = NULL, *ip = NULL, *password = NULL;
  int port = 0;
  CONFIGENTRY *tce;

  if (ce->ce_vardata != NULL)
  {
    /* Someone passed us a connect irc.hub.com { ... */
    name = ce->ce_vardata;
  }

  for (tce = ce->ce_entries; tce; tce = tce->ce_next)
  {
    if (!strcasecmp(tce->ce_varname, "NAME"))
    {
      if (tce->ce_vardata == NULL)
      {
        slog("%s:%i: no parameter for connect config option: %s",
             tce->ce_fileptr->cf_filename, tce->ce_varlinenum, tce->ce_varname);
        return 1;
      }
      name = tce->ce_vardata;
    }
    if (!strcasecmp(tce->ce_varname, "IP"))
    {
      if (tce->ce_vardata == NULL)
      {
        slog("%s:%i: no parameter for connect config option: %s",
             tce->ce_fileptr->cf_filename, tce->ce_varlinenum, tce->ce_varname);
        return 1;
      }
      ip = tce->ce_vardata;
    }
    else if (!strcasecmp(tce->ce_varname, "PASSWORD"))
    {
      if (tce->ce_vardata == NULL)
      {
        slog("%s:%i: no parameter for connect config option: %s",
             tce->ce_fileptr->cf_filename, tce->ce_varlinenum, tce->ce_varname);
        return 1;
      }
      password = tce->ce_vardata;
    }
    else if (!strcasecmp(tce->ce_varname, "PORT"))
    {
      if (tce->ce_vardata == NULL)
      {
        slog("%s:%i: no parameter for connect config option: %s",
             tce->ce_fileptr->cf_filename, tce->ce_varlinenum, tce->ce_varname);
        return 1;
      }
      port = tce->ce_vardatanum;
    }
  }
  if ((name != NULL) && (ip != NULL) && (password != NULL) && (port > 0) &&
      (port <= 65535))
  {
    add_server(ip, password, name, port);
    return 0;
  }
  /* else, we got an invalid conf block */
  slog("%s:%i: invalid connect block named: %s", ce->ce_fileptr->cf_filename,
       ce->ce_varlinenum, (name != NULL) ? name : "NONAME");
  return 1;
}

static int c_oper(CONFIGENTRY *ce)
{
  char *password = NULL, *name = NULL;
  u_short type = NOPRIV;
  CONFIGENTRY *tce;

  if (ce->ce_vardata != NULL)
  {
    /* Someone passed us an operator Hwy { ... */
    name = ce->ce_vardata;
  }

  for (tce = ce->ce_entries; tce; tce = tce->ce_next)
  {
    if (!strcasecmp(tce->ce_varname, "NAME"))
    {
      if (tce->ce_vardata == NULL)
      {
        slog("%s:%i: no parameter for operator config option: %s",
             tce->ce_fileptr->cf_filename, tce->ce_varlinenum, tce->ce_varname);
        return 1;
      }
      name = tce->ce_vardata;
    }
    else if (!strcasecmp(tce->ce_varname, "MASK"))
    {
      /* Do nothing, this value is ignored */
    }
    else if (!strcasecmp(tce->ce_varname, "PASSWORD"))
    {
      if (tce->ce_vardata == NULL)
      {
        slog("%s:%i: no parameter for operator config option: %s",
             tce->ce_fileptr->cf_filename, tce->ce_varlinenum, tce->ce_varname);
        return 1;
      }
      password = tce->ce_vardata;
    }
    else if (!strcasecmp(tce->ce_varname, "FLAG"))
    {
      if (tce->ce_vardata == NULL)
      {
        slog("%s:%i: no parameter for operator config option: %s",
             tce->ce_fileptr->cf_filename, tce->ce_varlinenum, tce->ce_varname);
        return 1;
      }
      if (!strcasecmp(tce->ce_vardata, "ADMIN"))
      {
        type = ADMIN;
      }
      else if (!strcasecmp(tce->ce_vardata, "CADMIN"))
      {
        type = CADMIN;
      }
      else if (!strcasecmp(tce->ce_vardata, "OPER"))
      {
        /* No special flags */
        type = SVSOP;
      }
      else
      {
        slog("%s:%i: Invalid operator type: %s", tce->ce_fileptr->cf_filename,
             tce->ce_varlinenum, tce->ce_vardata);
        return 1;
      }
    }
  }
  if ((name != NULL) && (password != NULL) && (type != NOPRIV))
  {
    add_oline(name, password, type);
    return 0;
  }
  /* else, we got an invalid conf block */
  slog("%s:%i: invalid operator block named: %s", ce->ce_fileptr->cf_filename,
       ce->ce_varlinenum, (name != NULL) ? name : "NONAME");
  return 1;
}

static int c_serverinfo(CONFIGENTRY *ce)
{
  generic_subblock_handler(ce, "SERVERINFO", conf_serverinfo_table);
  return 0;
}

static int c_settings(CONFIGENTRY *ce)
{
  generic_subblock_handler(ce, "SETTINGS", conf_settings_table);
  return 0;
}

static int c_channels(CONFIGENTRY *ce)
{
  int ccount = 0;

  if (ce->ce_entries == NULL)
  {
    slog("%s:%i: no entries for channels block",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
    return 1;
  }

  for (ce = ce->ce_entries;ce;ce = ce->ce_next)
  {
    if (ccount < MAXCHANS)
    {
      me.chans[ccount] = leetstrdup(ce->ce_varname);
      ccount++;
    }
  }
  me.chans[ccount] = NULL;
  return 0;
}

static int c_reservednick(CONFIGENTRY *ce)
{
  dlink_node *dl;

  if (ce->ce_entries == NULL)
  {
    slog("%s:%i: no entries for reservednick block",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
    return 1;
  }

  for (ce = ce->ce_entries;ce;ce = ce->ce_next)
  {
    dl = dlink_create();
    dlink_add(leetstrdup(ce->ce_varname), dl, &reserved_masks);
  }
  return 0;
}

static int c_services(CONFIGENTRY *ce)
{
  if (ce->ce_entries == NULL)
  {
    slog("%s:%i: no entries for services block",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
    return 1;
  }

  for (ce = ce->ce_entries;ce;ce = ce->ce_next)
  {
    add_services(ce->ce_varname);
  }
  return 0;
}

static int c_protected(CONFIGENTRY *ce)
{
  char *args, *nick, *user, *host, *gecos;

  if (ce->ce_entries == NULL)
  {
    slog("%s:%i: no entries for protected block",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
    return 1;
  }

  for (ce = ce->ce_entries;ce;ce = ce->ce_next)
  {
    args = ce->ce_varname;
    while (*args == ' ')
      args++; /* Ignore any leading spaces */
    nick = strtok(args, "!");
    user = strtok(NULL, "@");
    host = strtok(NULL, "$");
    gecos = strtok(NULL, " "); /* Ignore any trailing spaces */

    if ((nick == NULL) || (user == NULL) || (host == NULL) || (gecos == NULL))
    {
      slog("%s:%i: invalid entry in protected block",
           ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
      return 1;
    }

    add_protected(nick, user, host, gecos);
  }
  return 0;
}

static int c_operserv(CONFIGENTRY *ce)
{
  generic_subblock_handler(ce, "OPERSERV", conf_operserv_table);
  return 0;
}

static int c_si_name(CONFIGENTRY *ce)
{
  if (ce->ce_vardata == NULL)
  {
    slog("%s:%i: no parameter for serverinfo config option: %s",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
    return 1;
  }
  me.servname = leetstrdup(ce->ce_vardata);
  return 0;
}

static int c_si_ip(CONFIGENTRY *ce)
{
  if (ce->ce_vardata == NULL)
  {
    slog("%s:%i: no parameter for serverinfo config option: %s",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
    return 1;
  }
  me.ip.s_addr = inet_addr(ce->ce_vardata);
  return 0;
}

static int c_si_description(CONFIGENTRY *ce)
{
  if (ce->ce_vardata == NULL)
  {
    slog("%s:%i: no parameter for serverinfo config option: %s",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
    return 1;
  }
  me.servdesc = leetstrdup(ce->ce_vardata);
  return 0;
}

static int c_os_nick(CONFIGENTRY *ce)
{
  if (ce->ce_vardata == NULL)
  {
    slog("%s:%i: no parameter for operserv config option: %s",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
    return 1;
  }
  me.os_nick = leetstrdup(ce->ce_vardata);
  return 0;
}

static int c_os_username(CONFIGENTRY *ce)
{
  if (ce->ce_vardata == NULL)
  {
    slog("%s:%i: no parameter for operserv config option: %s",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
    return 1;
  }
  me.os_username = leetstrdup(ce->ce_vardata);
  return 0;
}

static int c_os_gecos(CONFIGENTRY *ce)
{
  if (ce->ce_vardata == NULL)
  {
    slog("%s:%i: no parameter for operserv config option: %s",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
    return 1;
  }
  me.os_gecos = leetstrdup(ce->ce_vardata);
  return 0;
}

static int c_se_opernotice(CONFIGENTRY *ce)
{
  me.opernotice = 1;
  return 0;
}

static int c_se_use_settime(CONFIGENTRY *ce)
{
  me.use_settime = 1;
  return 0;
}

static int c_se_operoperwall(CONFIGENTRY *ce)
{
  me.oper_operwall = 1;
  return 0;
}

static int c_se_ircdtype(CONFIGENTRY *ce)
{
  if (ce->ce_vardata == NULL)
  {
    slog("%s:%i: no parameter for settings config option: %s",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
    return 1;
  }
  if (!strcasecmp(ce->ce_vardata, "IRCNN"))
  {
    me.servtype = SERV_IRCNN;
  }
  else if (!strcasecmp(ce->ce_vardata, "HYBRID"))
  {
    me.servtype = SERV_HYBRD;
  }
  else
  {
    slog("%s:%i: invalid parameter for settings config option: %s: %s",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname,
         ce->ce_vardata);
    return 1;
  }
  return 0;
}

static int c_se_nicklen(CONFIGENTRY *ce) /* make network nicklen a conf option - jrb */
{
  if (ce->ce_vardata == NULL)
  {
    slog("%s:%i: no parameter for seettings config option: %s",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
    return 1;
  }
  me.nicklen = ce->ce_vardatanum;
  return 0;
}

static int c_se_conntimeout(CONFIGENTRY *ce)
{
  if (ce->ce_vardata == NULL)
  {
    slog("%s:%i: no parameter for seettings config option: %s",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
    return 1;
  }
  me.conntimeout = ce->ce_vardatanum;
  return 0;
}

static int c_se_authtimeout(CONFIGENTRY *ce)
{
  if (ce->ce_vardata == NULL)
  {
    slog("%s:%i: no parameter for seettings config option: %s",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
    return 1;
  }
  me.authtimeout = ce->ce_vardatanum;
  return 0;
}

static int c_se_clone_limit(CONFIGENTRY *ce)
{
  if (ce->ce_vardata == NULL)
  { 
    slog("%s:%i: no parameter for seettings config option: %s",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
    return 1;
  }
  me.clone_limit = ce->ce_vardatanum;
  return 0;
}

static int c_se_clone_gline(CONFIGENTRY *ce)
{
  if (ce->ce_vardata == NULL)
  {
    slog("%s:%i: no parameter for seettings config option: %s",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
    return 1;
  }
  me.clone_gline = ce->ce_vardatanum;
  return 0;
}

static int c_se_clone_gline_time(CONFIGENTRY *ce)
{ 
  if (ce->ce_vardata == NULL)
  { 
    slog("%s:%i: no parameter for seettings config option: %s",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
    return 1;
  }
  me.clone_gline_time = ce->ce_vardatanum;
  return 0;
}

static int c_se_clone_over(CONFIGENTRY *ce)
{ 
  if (ce->ce_vardata == NULL)
  { 
    slog("%s:%i: no parameter for seettings config option: %s",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
    return 1;
  }
  me.clone_over = ce->ce_vardatanum;
  return 0;
}

static int c_se_fclone_gline_time(CONFIGENTRY *ce)
{ 
  if (ce->ce_vardata == NULL)
  { 
    slog("%s:%i: no parameter for seettings config option: %s",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
    return 1;
  }
  me.fclone_gline_time = ce->ce_vardatanum;
  return 0;
}

static int c_se_maxkill(CONFIGENTRY *ce)
{ 
  if (ce->ce_vardata == NULL)
  { 
    slog("%s:%i: no parameter for seettings config option: %s",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
    return 1;
  }
  me.maxkill = ce->ce_vardatanum;
  return 0;
}

static int c_se_default_echo(CONFIGENTRY *ce)
{
  if (ce->ce_vardata == NULL)
  {
    slog("%s:%i: no parameter for seettings config option: %s",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
    return 1;
  }
  if (strcasecmp(ce->ce_vardata, "ON") == 0)
  {
    me.default_echo = 1;
  }
  else if (strcasecmp(ce->ce_vardata, "OFF") == 0)
  {
    me.default_echo = 0; /* No real effect */
  }
  else
  {
    slog("%s:%i: invalid parameter for seettings config option: %s",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
    return 1;
  }
  return 0;
}

static int c_se_floodrate(CONFIGENTRY *ce)
{
  if (ce->ce_vardata == NULL)
  {
    slog("%s:%i: no parameter for seettings config option: %s",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
    return 1;
  }
  me.floodrate = ce->ce_vardatanum;
  return 0;
}

static int c_se_cloneurl(CONFIGENTRY *ce)
{
  if (ce->ce_vardata == NULL)
  {
    slog("%s:%i: no parameter for seettings config option: %s",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
    return 1;
  }
  if (me.cloneurl != NULL)
    leetfree(me.cloneurl, strlen(me.cloneurl));
  me.cloneurl = leetstrdup(ce->ce_vardata);
  return 0;
}

static int c_se_nickpass(CONFIGENTRY *ce)
{
  if (ce->ce_vardata == NULL)
  {
    slog("%s:%i: no parameter for seettings config option: %s",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
    return 1;
  }
  me.nickpass = leetstrdup(ce->ce_vardata);
  return 0;
}

static int c_se_nickserv(CONFIGENTRY *ce)
{
  if (ce->ce_vardata == NULL)
  {
    slog("%s:%i: no parameter for seettings config option: %s",
         ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
    return 1;
  }
  me.nickserv = leetstrdup(ce->ce_vardata);
  return 0;
}

void InitDefaults(void)
{
  me.cloneurl = leetstrdup("http://www.yourircnetworks.net");
  me.floodrate = 4;
  me.nicklen = 9;
  me.conntimeout = 30;
  me.authtimeout = 60;
  reserved_masks.head = NULL;
  reserved_masks.tail = NULL;
  me.clone_limit = 3;
  me.clone_gline = 8;
  me.clone_gline_time = 1;
  me.clone_over = 3;
  me.fclone_gline_time = 6;
  me.maxkill = 64;

  /* XXX etc */
  main_sock.head = NULL;
  main_sock.tail = NULL;
}

void VerifyConf(void)
{
  if (!me.servname || !me.servdesc)
  {
    fprintf(stderr, "serverinfo{} incomplete, fatal");
    exit(1);
  }

  if (!me.os_nick || check_nick(NULL, me.os_nick) || !me.os_username ||
      check_um(NULL, me.os_username) || !me.os_gecos ||
      check_rm(NULL, me.os_gecos))
  {
    fprintf(stderr, "operserv{} incomplete, fatal");
    exit(2);
  }

  if (!me.chans[0])
  {
    fprintf(stderr, "You must have one entry in your channels{}, fatal");
    exit(3);
  }

  if (me.clone_limit < 2)
  {
    slog("CONF_ERROR:  clone_limit may not be below 2, defaulting to 3.");
    me.clone_limit = 3;
  }

  if (me.clone_gline < 1)
  {
    slog("CONF_ERROR:  clone_gline may not be below 1, defaulting to 8.");
    me.clone_gline = 8;
  }

  if (me.clone_limit >= me.clone_gline)
  {
    me.clone_gline = me.clone_limit + 1;
    slog("CONF_ERROR:  clone_gline may not be below clone_limit, setting to %d.",
         me.clone_gline);
  }

  if (me.clone_gline_time < 1)
  {
    slog("CONF_ERROR:  clone_gline_time may not be below 1, defaulting to 1.");
    me.clone_gline_time = 1;
  }

  if (me.clone_over < 0)
  {
    slog("CONF_ERROR:  clone_over may not be below 0, defaulting to 3.");
    me.clone_over = 3;
  }

  if (me.fclone_gline_time < 1)
  {
    slog("CONF_ERROR:  fclone_gline_time may not be below 1, defaulting to 6.");
    me.fclone_gline_time = 6;
  }

  if (me.maxkill < 1)
  {
    slog("CONF_ERROR:  maxkill may not be below 1, defaulting to 64.");
    me.maxkill = 64;
  }
}

void rehash(void)
{
  CONFIGFILE *cfptr;
  CONFIGFILE *cfp;
  CONFIGENTRY *ce;
  struct ConfTable *ct = NULL;
  dlink_node *dl, *dtmp;

  tell_chans("Rehashing...");
  slog("rehashing...");

  /* Reset the boolean values to false */
  me.opernotice = 0;
  me.use_settime = 0;

  del_all_servers();
  del_all_olines();
  flush_protected();
  flush_services();

  DLINK_FOREACH_SAFE(dl, dtmp, reserved_masks.head)
  {
    dlink_delete(dl, &reserved_masks);
    free(dl->data);
    dlink_free(dl);
  }

  cfptr = cfp = config_load(CONF);
  if (cfp == NULL)
  {
    slog("Unable to open configuration file (%s): %s\n",
         CONF, strerror(errno));
  }
  for (;cfptr;cfptr = cfptr->cf_next) /* Handle include files? */
  {
    for (ce = cfptr->cf_entries;ce;ce = ce->ce_next)
    {
      for (ct = conf_root_table; ct->name; ct++)
      {
        if (!strcasecmp(ct->name, ce->ce_varname))
        {
          if (ct->rehashable == 1)
            ct->handler(ce);
          break;
        }
      }
    }
    if (ct->name == NULL)
    {
      slog("%s:%i: invalid config option: %s",
           ce->ce_fileptr->cf_filename, ce->ce_varlinenum,
           ce->ce_varname);
    }
  }
  config_free(cfp);

  if (me.logfd)
  {
    fflush(me.logfd);
    fclose(me.logfd);
  }
  if (!(me.logfd = fopen(LOGFILE, "a")))
  {
    tell_chans("\2[ERROR]\2 Unable to open logfile: %s", strerror(errno));
    /* slog("Unable to open logfile: %s", strerror(errno)); */
  }
  tell_chans("done!");
  slog("done!");
}
