/*
*  IRC Services
*
*  Copyright (C) 2003 Services Developers
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of version 1 of the GNU General Public License,
*  as published by the Free Software Foundation.
*
*  This program is distributed in the hope that it will be useful, but
*  WITHOUT ANY WARRANTY; without even the warranty of MERCHANTABILITY
*  or FITNESS FOR A PARTICULAR PURPOSE. The GNU General Public License
*  contains more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
* $Id: irc_string.c,v 1.28 2003/10/30 02:42:42 wcampbel Exp $
*/

#include <assert.h>

#include "irc_types.h"
#include "irc_string.h"
#include "defines.h"
#include "send.h"
#include "me.h"
#include "md5.h"

/*
 * strlcat and strlcpy extracted from ircd-hybrid, originally taken
 * from the FreeBSD source.  This copy fixes a bug.
 *
 * They had the following Copyright info:
 *
 *
 * Copyright (c) 1998 Todd C. Miller <Todd.Miller@courtesan.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED `AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef HAVE_STRLCAT
size_t strlcat(char *dst, const char *src, size_t siz)
{
  char *d = dst;
  const char *s = src;
  size_t n = siz, dlen;

  while (n-- != 0 && *d != '\0')
    d++;
  dlen = d - dst;
  n = siz - dlen;

  if (n == 0)
    return (dlen + strlen(s));
  while (*s != '\0')
  {
    if (n != 1)
    {
      *d++ = *s;
      n--;
    }
    s++;
  }
  *d = '\0';
  return (dlen + (s - src));       /* count does not include NUL */
}
#endif

#ifndef HAVE_STRLCPY
size_t strlcpy(char *dst, const char *src, size_t siz)
{
  char *d = dst;
  const char *s = src;
  size_t n = siz;
  /* Copy as many bytes as will fit */
  if (n != 0 && --n != 0)
  {
    do
    {
      if ((*d++ = *s++) == 0)
        break;
    }
    while (--n != 0);
  }
  /* Not enough room in dst, add NUL and traverse rest of src */
  if (n == 0)
  {
    if (siz != 0)
      *d = '\0';              /* NUL-terminate dst */
    while (*s++)
      ;
  }

  return (s - src - 1);    /* count does not include NUL */
}
#endif

char *
strnull(char *stuff)
{
  return (stuff ? stuff : "(null)");
}

/* match() originally written by Douglas A Lewis (dalewis@acsu.buffalo.edu) */
u_short
match(const char *mask, const char *name)
{
  const unsigned char *m = (const unsigned char *)mask;
  const unsigned char *n = (const unsigned char *)name;
  const unsigned char *ma = (const unsigned char *)mask;
  const unsigned char *na = (const unsigned char *)name;
  u_short wild = 0;

  assert(0 != mask);
  assert(0 != name);
loopy:
  if (*m == '*')
  {
    while (*m == '*')
      m++;
    wild = 1;
    ma = m;
    na = n;
  }
  if (!*m)
  {
    if (!*n)
      return 1;
    for (m--; (m > (const unsigned char *)mask) &&
               (*m == '?' || *m == '&' || *m == '#'); m--)
      ;
    if (*m == '*' && (m > (const unsigned char *) mask))
      return 1;
    if (!wild)
      return 0;
    m = ma;
    n = ++na;
  }
  else if (!*n)
  {
    while (*m == '*')
      m++;
    return (*m == 0);
  }
  if (ToLower(*m) != ToLower(*n) && !(*m == '?') &&
      !(*m == '&' && IsLetter(*n)) &&
      !(*m == '#' && IsDigit(*n)))
  {
    if (!wild)
      return 0;
    m = ma;
    n = ++na;
  }
  else
  {
    if (*m)
      m++;
    if (*n)
      n++;
  }
  goto loopy;
}

u_short
check_nick(user_t *cptr, char *nick)
{
  if (strlen(nick) > me.nicklen)
    return (cptr ? reply(cptr->nick, "The nickname is too long.") : 1);
  if (isdigit((int) *nick) || *nick == '-')
    return (cptr ? reply(cptr->nick, "Erroneus Nickname.") : 1);
  for (; *nick; nick++)
    if (!IsNickChar(*nick))
      return (cptr ? reply(cptr->nick, "The character '%c' is not allowed in your nick.", *nick) : 1);
  return 0;
}

u_short
check_nick_sock(sock_t *sock, char *nick)
{
  if (strlen(nick) > me.nicklen)
  {
    tosock(sock, "The nickname is too long.");
    return 1;
  }

  if (isdigit((int) *nick) || *nick == '-')
  {
    tosock(sock, "Erroneus Nickname.");
    return 1;
  }

  for (; *nick; nick++)
  {
    if (!IsNickChar(*nick))
    {
      tosock(sock, "The character '%c' is not allowed in your nick.", *nick);
      return 1;
    }
  }

  return 0;
}

u_short
check_nm_sock(sock_t *sock, char *nm)
{
  if (strlen(nm) > me.nicklen)
  {
    tosock(sock, "The nickmask is too long.");
    return 1;
  }
  for (; *nm; nm++)
    if (!IsNickChar(*nm) && !IsMatchC(*nm))
    {
      tosock(sock, "The character '%c' is not allowed in your nickmask.", *nm);
      return 1;
    }
  return 0;
}

u_short
check_um(user_t *cptr, char *um)
{
  if (strlen(um) > USERLEN)
    return (cptr ? reply(cptr->nick, "The usermask is too long.") : 1);
  for (; *um; um++)
    if (!IsUserChar(*um) && !IsMatchC(*um))
      return (cptr ? reply(cptr->nick, "The character '%c' is not allowed in your usermask.", *um) : 1);
  return 0;
}

u_short
check_um_sock(sock_t *sock, char *um)
{
  if (strlen(um) > USERLEN)
  {
    tosock(sock, "The usermask is too long.");
    return 1;
  }
  for (; *um; um++)
    if (!IsUserChar(*um) && !IsMatchC(*um))
    {
      tosock(sock, "The character '%c' is not allowed in your usermask.", *um);
      return 1;
    }
  return 0;
}

u_short
check_host(user_t *cptr, char *hm)
{
  if (strlen(hm) > HOSTLEN)
    return (cptr ? reply(cptr->nick, "Your host is too long.") : 1);
  if (!strchr(hm, '.'))
    return (cptr ? reply(cptr->nick, "Your host must contain at least one period.") : 1);
  for (; *hm; hm++)
    if (!IsHostChar(*hm))
      return (cptr ? reply(cptr->nick, "The %s '%c' is not allowed in your hostmask.", ((IsMatchC(*hm)) ? "wildcard" : "character"), *hm) : 1);
  return 0;
}

u_short
check_host_sock(sock_t *sock, char *hm)
{
  if (strlen(hm) > HOSTLEN)
  {
    tosock(sock, "Your host is too long.");
    return 1;
  }
  if (!strchr(hm, '.'))
  {
    tosock(sock, "Your host must contain at least one period.");
    return 1;
  }
  for (; *hm; hm++)
    if (!IsHostChar(*hm))
    {
      tosock(sock, "The %s '%c' is not allowed in your hostmask.", ((IsMatchC(*hm)) ? "wildcard" : "character"), *hm);
      return 1;
    }
  return 0;
}

u_short
check_hm(user_t *cptr, char *hm)
{
  if (strlen(hm) > HOSTLEN)
    return (cptr ? reply(cptr->nick, "The hostmask is too long.") : 1);
  for (; *hm; hm++)
    if (!IsHostChar(*hm) && !IsMatchC(*hm))
      return (cptr ? reply(cptr->nick, "The character '%c' is not allowed in your hostmask.", *hm) : 1);
  return 0;
}

u_short
check_hm_sock(sock_t *sock, char *hm)
{
  if (strlen(hm) > HOSTLEN)
  {
    tosock(sock, "The hostmask is too long.");
    return 1;
  }
  for (; *hm; hm++)
    if (!IsHostChar(*hm) && !IsMatchC(*hm))
    {
      tosock(sock, "The character '%c' is not allowed in your hostmask.", *hm);
      return 1;
    }
  return 0;
}

u_short
check_rm(user_t *cptr, char *rm)
{
  if (strlen(rm) > REALLEN)
    return (cptr ? reply(cptr->nick, "The realname mask is too long.") : 1);
  for (; *rm; rm++)
    if (*rm == ':')
      return (cptr ? reply(cptr->nick, "The character '%c' is not allowed in your realname mask.", *rm) : 1);
  return 0;
}

u_short
check_rm_sock(sock_t *sock, char *rm)
{
  if (strlen(rm) > REALLEN)
  {
    tosock(sock, "The realname mask is too long.");
    return 1;
  }
  for (; *rm; rm++)
    if (*rm == ':')
    {
      tosock(sock, "The character '%c' is not allowed in your realname mask.", *rm);
      return 1;
    }
  return 0;
}

u_short
check_chan(sock_t *sock, char *chan)
{
  if (strlen(chan) > CHANLEN)
  {
    tosock(sock, "That is too long for a channel name.");
    return 1;
  }
  if (!IsChanPrefix(*chan))
  {
    tosock(sock, "That is not a valid channel name.");
    return 1;
  }
  for (; *chan; chan++)
    if (!IsChanChar(*chan))
    {
      tosock(sock, "The character '%c' is not allowed in a channel name.", *chan);
      return 1;
    }
  return 0;
}


u_short
check_reason(user_t *cptr, char *reason)
{
  if (strlen(reason) > KILLLEN)
    return (cptr ? reply(cptr->nick, "The reason is too long.") : 1);
  for (; *reason; reason++)
    if (*reason == ':' || *reason == '#')
      return (cptr ? reply(cptr->nick, "The character '%c' is not allowed in your reason.", *reason) : 1);
  return 0;
}

u_short
check_reason_sock(sock_t *sock, char *reason)
{
  if (strlen(reason) > KILLLEN)
  {
    tosock(sock, "The reason is too long.");
    return 1;
  }
  for (; *reason; reason++)
    if (*reason == ':' || *reason == '#')
    {
      tosock(sock, "The character '%c' is not allowed in your reason.", *reason);
      return 1;
    }
  return 0;
}

void
replace_colon(char *in)
{
  for (; *in; in++)
    if (*in == ':')
      *in = ';';
}

char *
leetctime(time_t lclock)
{
  static char buf[25];

  strlcpy(buf, ctime(&lclock), 25);
  return buf;
}

char *
shtime(void)
{
  static char buf[9];
#ifdef SECONDS_IN_TIMESTAMP

  strlcpy(buf, leetctime(time(NULL)) + 11, 9);
#else

  strlcpy(buf, leetctime(time(NULL)) + 11, 6);
#endif

  return buf;
}

const unsigned int validchars[] =
  {
    /* 0  */     CNTRL_C,
    /* 1  */     CNTRL_C |  CHAN_C |  NONEOS_C,
    /* 2  */     CNTRL_C |  CHAN_C |  NONEOS_C,
    /* 3  */     CNTRL_C |  CHAN_C |  NONEOS_C,
    /* 4  */     CNTRL_C |  CHAN_C |  NONEOS_C,
    /* 5  */     CNTRL_C |  CHAN_C |  NONEOS_C,
    /* 6  */     CNTRL_C |  CHAN_C |  NONEOS_C,
    /* 7 BEL */  CNTRL_C |  NONEOS_C,
    /* 8  \b */  CNTRL_C |  CHAN_C |  NONEOS_C,
    /* 9  \t */  CNTRL_C |  SPACE_C |  CHAN_C |  NONEOS_C,
    /* 10 \n */  CNTRL_C |  SPACE_C |  CHAN_C |  NONEOS_C |  EOL_C,
    /* 11 \v */  CNTRL_C |  SPACE_C |  CHAN_C |  NONEOS_C,
    /* 12 \f */  CNTRL_C |  SPACE_C |  CHAN_C |  NONEOS_C,
    /* 13 \r */  CNTRL_C |  SPACE_C |  CHAN_C |  NONEOS_C |  EOL_C,
    /* 14 */     CNTRL_C |  CHAN_C |  NONEOS_C,
    /* 15 */     CNTRL_C |  CHAN_C |  NONEOS_C,
    /* 16 */     CNTRL_C |  CHAN_C |  NONEOS_C,
    /* 17 */     CNTRL_C |  CHAN_C |  NONEOS_C,
    /* 18 */     CNTRL_C |  CHAN_C |  NONEOS_C,
    /* 19 */     CNTRL_C |  CHAN_C |  NONEOS_C,
    /* 20 */     CNTRL_C |  CHAN_C |  NONEOS_C,
    /* 21 */     CNTRL_C |  CHAN_C |  NONEOS_C,
    /* 22 */     CNTRL_C |  CHAN_C |  NONEOS_C,
    /* 23 */     CNTRL_C |  CHAN_C |  NONEOS_C,
    /* 24 */     CNTRL_C |  CHAN_C |  NONEOS_C,
    /* 25 */     CNTRL_C |  CHAN_C |  NONEOS_C,
    /* 26 */     CNTRL_C |  CHAN_C |  NONEOS_C,
    /* 27 */     CNTRL_C |  CHAN_C |  NONEOS_C,
    /* 28 */     CNTRL_C |  CHAN_C |  NONEOS_C,
    /* 29 */     CNTRL_C |  CHAN_C |  NONEOS_C,
    /* 30 */     CNTRL_C |  CHAN_C |  NONEOS_C,
    /* 31 */     CNTRL_C |  CHAN_C |  NONEOS_C,
    /* SP */     PRINT_C |  SPACE_C,
    /* ! */      PRINT_C |  KWILD_C |  CHAN_C |  NONEOS_C,
    /* " */      PRINT_C |  CHAN_C |  NONEOS_C,
    /* # */      PRINT_C |  CHANPFX_C |  CHAN_C |  NONEOS_C  | MATCH_C,
    /* $ */      PRINT_C |  CHAN_C |  NONEOS_C |  USER_C,
    /* % */      PRINT_C |  CHAN_C |  NONEOS_C,
    /* & */      PRINT_C |  CHANPFX_C |  CHAN_C |  NONEOS_C  | MATCH_C,
    /* ' */      PRINT_C |  CHAN_C |  NONEOS_C,
    /* ( */      PRINT_C |  CHAN_C |  NONEOS_C,
    /* ) */      PRINT_C |  CHAN_C |  NONEOS_C,
    /* * */      PRINT_C |  KWILD_C |  CHAN_C |  NONEOS_C |  SERV_C | MATCH_C,
    /* + */      PRINT_C |  CHAN_C |  NONEOS_C,
    /* , */      PRINT_C |  NONEOS_C,
    /* - */      PRINT_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C,
    /* . */      PRINT_C |  KWILD_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C |  SERV_C,
    /* / */      PRINT_C |  CHAN_C |  NONEOS_C,
    /* 0 */      PRINT_C |  DIGIT_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C,
    /* 1 */      PRINT_C |  DIGIT_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C,
    /* 2 */      PRINT_C |  DIGIT_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C,
    /* 3 */      PRINT_C |  DIGIT_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C,
    /* 4 */      PRINT_C |  DIGIT_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C,
    /* 5 */      PRINT_C |  DIGIT_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C,
    /* 6 */      PRINT_C |  DIGIT_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C,
    /* 7 */      PRINT_C |  DIGIT_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C,
    /* 8 */      PRINT_C |  DIGIT_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C,
    /* 9 */      PRINT_C |  DIGIT_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C,
    /* : */      PRINT_C |  CHAN_C |  NONEOS_C |  HOST_C,
    /* ; */      PRINT_C |  CHAN_C |  NONEOS_C,
    /* < */      PRINT_C |  CHAN_C |  NONEOS_C,
    /* = */      PRINT_C |  CHAN_C |  NONEOS_C,
    /* > */      PRINT_C |  CHAN_C |  NONEOS_C,
    /* ? */      PRINT_C |  KWILD_C |  CHAN_C |  NONEOS_C  | MATCH_C,
    /* @ */      PRINT_C |  KWILD_C |  CHAN_C |  NONEOS_C,
    /* A */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* B */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* C */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* D */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* E */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* F */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* G */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* H */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* I */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* J */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* K */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* L */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* M */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* N */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* O */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* P */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* Q */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* R */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* S */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* T */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* U */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* V */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* W */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* X */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* Y */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* Z */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* [ */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C,
    /* \ */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C,
    /* ] */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C,
    /* ^ */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C,
    /* _ */      PRINT_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C,
    /* ` */      PRINT_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C,
    /* a */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* b */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* c */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* d */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* e */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* f */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* g */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* h */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* i */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* j */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* k */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* l */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* m */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* n */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* o */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* p */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* q */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* r */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* s */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* t */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* u */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* v */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* w */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* x */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* y */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* z */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C |  HOST_C | LET_C,
    /* { */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C,
    /* | */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C,
    /* } */      PRINT_C |  ALPHA_C |  NICK_C |  CHAN_C |  NONEOS_C |  USER_C,
    /* ~ */      PRINT_C |  ALPHA_C |  CHAN_C |  NONEOS_C |  USER_C,
    /* del  */   CHAN_C |  NONEOS_C,
    /* 0x80 */   CHAN_C |  NONEOS_C,
    /* 0x81 */   CHAN_C |  NONEOS_C,
    /* 0x82 */   CHAN_C |  NONEOS_C,
    /* 0x83 */   CHAN_C |  NONEOS_C,
    /* 0x84 */   CHAN_C |  NONEOS_C,
    /* 0x85 */   CHAN_C |  NONEOS_C,
    /* 0x86 */   CHAN_C |  NONEOS_C,
    /* 0x87 */   CHAN_C |  NONEOS_C,
    /* 0x88 */   CHAN_C |  NONEOS_C,
    /* 0x89 */   CHAN_C |  NONEOS_C,
    /* 0x8A */   CHAN_C |  NONEOS_C,
    /* 0x8B */   CHAN_C |  NONEOS_C,
    /* 0x8C */   CHAN_C |  NONEOS_C,
    /* 0x8D */   CHAN_C |  NONEOS_C,
    /* 0x8E */   CHAN_C |  NONEOS_C,
    /* 0x8F */   CHAN_C |  NONEOS_C,
    /* 0x90 */   CHAN_C |  NONEOS_C,
    /* 0x91 */   CHAN_C |  NONEOS_C,
    /* 0x92 */   CHAN_C |  NONEOS_C,
    /* 0x93 */   CHAN_C |  NONEOS_C,
    /* 0x94 */   CHAN_C |  NONEOS_C,
    /* 0x95 */   CHAN_C |  NONEOS_C,
    /* 0x96 */   CHAN_C |  NONEOS_C,
    /* 0x97 */   CHAN_C |  NONEOS_C,
    /* 0x98 */   CHAN_C |  NONEOS_C,
    /* 0x99 */   CHAN_C |  NONEOS_C,
    /* 0x9A */   CHAN_C |  NONEOS_C,
    /* 0x9B */   CHAN_C |  NONEOS_C,
    /* 0x9C */   CHAN_C |  NONEOS_C,
    /* 0x9D */   CHAN_C |  NONEOS_C,
    /* 0x9E */   CHAN_C |  NONEOS_C,
    /* 0x9F */   CHAN_C |  NONEOS_C,
    /* 0xA0 */   CHAN_C |  NONEOS_C,
    /* 0xA1 */   CHAN_C |  NONEOS_C,
    /* 0xA2 */   CHAN_C |  NONEOS_C,
    /* 0xA3 */   CHAN_C |  NONEOS_C,
    /* 0xA4 */   CHAN_C |  NONEOS_C,
    /* 0xA5 */   CHAN_C |  NONEOS_C,
    /* 0xA6 */   CHAN_C |  NONEOS_C,
    /* 0xA7 */   CHAN_C |  NONEOS_C,
    /* 0xA8 */   CHAN_C |  NONEOS_C,
    /* 0xA9 */   CHAN_C |  NONEOS_C,
    /* 0xAA */   CHAN_C |  NONEOS_C,
    /* 0xAB */   CHAN_C |  NONEOS_C,
    /* 0xAC */   CHAN_C |  NONEOS_C,
    /* 0xAD */   CHAN_C |  NONEOS_C,
    /* 0xAE */   CHAN_C |  NONEOS_C,
    /* 0xAF */   CHAN_C |  NONEOS_C,
    /* 0xB0 */   CHAN_C |  NONEOS_C,
    /* 0xB1 */   CHAN_C |  NONEOS_C,
    /* 0xB2 */   CHAN_C |  NONEOS_C,
    /* 0xB3 */   CHAN_C |  NONEOS_C,
    /* 0xB4 */   CHAN_C |  NONEOS_C,
    /* 0xB5 */   CHAN_C |  NONEOS_C,
    /* 0xB6 */   CHAN_C |  NONEOS_C,
    /* 0xB7 */   CHAN_C |  NONEOS_C,
    /* 0xB8 */   CHAN_C |  NONEOS_C,
    /* 0xB9 */   CHAN_C |  NONEOS_C,
    /* 0xBA */   CHAN_C |  NONEOS_C,
    /* 0xBB */   CHAN_C |  NONEOS_C,
    /* 0xBC */   CHAN_C |  NONEOS_C,
    /* 0xBD */   CHAN_C |  NONEOS_C,
    /* 0xBE */   CHAN_C |  NONEOS_C,
    /* 0xBF */   CHAN_C |  NONEOS_C,
    /* 0xC0 */   CHAN_C |  NONEOS_C,
    /* 0xC1 */   CHAN_C |  NONEOS_C,
    /* 0xC2 */   CHAN_C |  NONEOS_C,
    /* 0xC3 */   CHAN_C |  NONEOS_C,
    /* 0xC4 */   CHAN_C |  NONEOS_C,
    /* 0xC5 */   CHAN_C |  NONEOS_C,
    /* 0xC6 */   CHAN_C |  NONEOS_C,
    /* 0xC7 */   CHAN_C |  NONEOS_C,
    /* 0xC8 */   CHAN_C |  NONEOS_C,
    /* 0xC9 */   CHAN_C |  NONEOS_C,
    /* 0xCA */   CHAN_C |  NONEOS_C,
    /* 0xCB */   CHAN_C |  NONEOS_C,
    /* 0xCC */   CHAN_C |  NONEOS_C,
    /* 0xCD */   CHAN_C |  NONEOS_C,
    /* 0xCE */   CHAN_C |  NONEOS_C,
    /* 0xCF */   CHAN_C |  NONEOS_C,
    /* 0xD0 */   CHAN_C |  NONEOS_C,
    /* 0xD1 */   CHAN_C |  NONEOS_C,
    /* 0xD2 */   CHAN_C |  NONEOS_C,
    /* 0xD3 */   CHAN_C |  NONEOS_C,
    /* 0xD4 */   CHAN_C |  NONEOS_C,
    /* 0xD5 */   CHAN_C |  NONEOS_C,
    /* 0xD6 */   CHAN_C |  NONEOS_C,
    /* 0xD7 */   CHAN_C |  NONEOS_C,
    /* 0xD8 */   CHAN_C |  NONEOS_C,
    /* 0xD9 */   CHAN_C |  NONEOS_C,
    /* 0xDA */   CHAN_C |  NONEOS_C,
    /* 0xDB */   CHAN_C |  NONEOS_C,
    /* 0xDC */   CHAN_C |  NONEOS_C,
    /* 0xDD */   CHAN_C |  NONEOS_C,
    /* 0xDE */   CHAN_C |  NONEOS_C,
    /* 0xDF */   CHAN_C |  NONEOS_C,
    /* 0xE0 */   CHAN_C |  NONEOS_C,
    /* 0xE1 */   CHAN_C |  NONEOS_C,
    /* 0xE2 */   CHAN_C |  NONEOS_C,
    /* 0xE3 */   CHAN_C |  NONEOS_C,
    /* 0xE4 */   CHAN_C |  NONEOS_C,
    /* 0xE5 */   CHAN_C |  NONEOS_C,
    /* 0xE6 */   CHAN_C |  NONEOS_C,
    /* 0xE7 */   CHAN_C |  NONEOS_C,
    /* 0xE8 */   CHAN_C |  NONEOS_C,
    /* 0xE9 */   CHAN_C |  NONEOS_C,
    /* 0xEA */   CHAN_C |  NONEOS_C,
    /* 0xEB */   CHAN_C |  NONEOS_C,
    /* 0xEC */   CHAN_C |  NONEOS_C,
    /* 0xED */   CHAN_C |  NONEOS_C,
    /* 0xEE */   CHAN_C |  NONEOS_C,
    /* 0xEF */   CHAN_C |  NONEOS_C,
    /* 0xF0 */   CHAN_C |  NONEOS_C,
    /* 0xF1 */   CHAN_C |  NONEOS_C,
    /* 0xF2 */   CHAN_C |  NONEOS_C,
    /* 0xF3 */   CHAN_C |  NONEOS_C,
    /* 0xF4 */   CHAN_C |  NONEOS_C,
    /* 0xF5 */   CHAN_C |  NONEOS_C,
    /* 0xF6 */   CHAN_C |  NONEOS_C,
    /* 0xF7 */   CHAN_C |  NONEOS_C,
    /* 0xF8 */   CHAN_C |  NONEOS_C,
    /* 0xF9 */   CHAN_C |  NONEOS_C,
    /* 0xFA */   CHAN_C |  NONEOS_C,
    /* 0xFB */   CHAN_C |  NONEOS_C,
    /* 0xFC */   CHAN_C |  NONEOS_C,
    /* 0xFD */   CHAN_C |  NONEOS_C,
    /* 0xFE */   CHAN_C |  NONEOS_C,
    /* 0xFF */   CHAN_C |  NONEOS_C
  };

const unsigned char ToLowerTab[] =
  {
    0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xa,
    0xb, 0xc, 0xd, 0xe, 0xf, 0x10, 0x11, 0x12, 0x13, 0x14,
    0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d,
    0x1e, 0x1f,
    ' ', '!', '"', '#', '$', '%', '&', 0x27, '(', ')',
    '*', '+', ',', '-', '.', '/',
    '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
    ':', ';', '<', '=', '>', '?',
    '@', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i',
    'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's',
    't', 'u', 'v', 'w', 'x', 'y', 'z', '{', '|', '}', '~',
    '_',
    '`', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i',
    'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's',
    't', 'u', 'v', 'w', 'x', 'y', 'z', '{', '|', '}', '~',
    0x7f,
    0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
    0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
    0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99,
    0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
    0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9,
    0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf,
    0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9,
    0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf,
    0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9,
    0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
    0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9,
    0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
    0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9,
    0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
    0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9,
    0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff
  };

const unsigned char ToUpperTab[] =
  {
    0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xa,
    0xb, 0xc, 0xd, 0xe, 0xf, 0x10, 0x11, 0x12, 0x13, 0x14,
    0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d,
    0x1e, 0x1f,
    ' ', '!', '"', '#', '$', '%', '&', 0x27, '(', ')',
    '*', '+', ',', '-', '.', '/',
    '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
    ':', ';', '<', '=', '>', '?',
    '@', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I',
    'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S',
    'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '[', '\\', ']', '^',
    0x5f,
    '`', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I',
    'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S',
    'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '[', '\\', ']', '^',
    0x7f,
    0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
    0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
    0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99,
    0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
    0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9,
    0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf,
    0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9,
    0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf,
    0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9,
    0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
    0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9,
    0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
    0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9,
    0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
    0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9,
    0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff
  };

char *
wiz_md5(char *stuff)
{
  MD5_CTX context;
  u_char digest[16];
  u_short i = 0, j = 0;

  MD5Init(&context);
  MD5Update(&context, (unsigned char *)stuff, strlen(stuff));
  MD5Final(digest, &context);

  me.md5buf[0] = '\0';
  for (i = 0, j = 0; i < 16; i++, j += 2)
    sprintf(&me.md5buf[j], "%02x", digest[i]);
  me.md5buf[32] = '\0';

  return me.md5buf;
}

