/*
*  IRC Services
*
*  Copyright (C) 2003 Services Developers
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of version 1 of the GNU General Public License,
*  as published by the Free Software Foundation.
*
*  This program is distributed in the hope that it will be useful, but
*  WITHOUT ANY WARRANTY; without even the warranty of MERCHANTABILITY
*  or FITNESS FOR A PARTICULAR PURPOSE. The GNU General Public License
*  contains more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
* $Id: main.c,v 1.52 2003/11/05 15:19:50 wcampbel Exp $
*/

#include "irc_types.h"
#include "irc_string.h"
#include "defines.h"
#include "events.h"
#include "config.h"
#include "main.h"
#include "send.h"
#include "mem.h"
#define OPEN_DATABASES
#include "db.h"
#undef OPEN_DATABASES
#include "me.h"
#include "servers.h"
#include "clone.h"
#include "oline.h"
#include "snuggle.h"
#include "conf.h"
#include "bconf.h"
#include "channels.h"
#include "net.h"
#include "dcc.h"
#include "serno.h"
#include "csnuggle.h" /* needed for del_all_csnuggles() */

#include <sys/stat.h>
#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/resource.h>

#ifdef MEMDEBUG
char *malloc_options = "AJX";
#endif

struct _me me;

static void setup_corefile(void);
static void check_midnight(void);
static int is_midnight(void);

u_short
slog(char *input, ...)
{
  char logbuf[BUFSIZE];
  va_list ap;

  if (!me.logfd)
  {
    tell_chans("\2[ERROR]\2 Unable to open logfile!");
    return 1;
  }
  va_start(ap, input);
  vsnprintf(logbuf, BUFSIZE, input, ap);
  fprintf(me.logfd, "%s %s\n", leetctime(time(NULL)), logbuf);
  fflush(me.logfd);

  if (me.debug)
    printf("LOG %s\n", logbuf);

  va_end(ap);
  return 1;
}

u_short
reload(void)
{
  u_short i = 0;

  tell_chans("Reloading my databases...");
  slog("reloading databases...");
  del_all_clone_allows();
  del_all_snuggles();
  del_all_csnuggles(); /* Be sure to clear these out too */
  if ((i = read_db(0)) > 0)
  {
    tell_chans("\2[ERROR]\2 Unable to read %s: %s", databases[i], strerror(errno));
    goto r_err;
  }
  tell_chans("done!");
  slog("done!");
  return 0;
r_err:
  return tell_chans("\2[ERROR]\2 Error while reloading!");
}

int
main(int argc, char **argv)
{
  int x;
  FILE *pidf;
  int no_fork = 0;

  umask(0077);

  /* initialize stuff */
  memset(&me, 0, sizeof(me));
  me.argc = argc;
  me.argv = argv;
  me.start = time(NULL);
  me.settime = time(NULL) + (60 * 60 * 4);

  setup_corefile();

  /* setup signals */
  signal(SIGHUP, do_signal);
  signal(SIGINT, do_signal);
  signal(SIGQUIT, do_signal);
  signal(SIGTRAP, SIG_IGN);
#if 0

  signal(SIGILL, do_signal);
  signal(SIGBUS, do_signal);
  signal(SIGSEGV, do_signal);
  signal(SIGSYS, do_signal);
#endif

  signal(SIGALRM, SIG_IGN);
  signal(SIGTERM, do_signal);
  signal(SIGPIPE, SIG_IGN);
  signal(SIGUSR1, SIG_IGN);
  signal(SIGUSR2, SIG_IGN);

  while ((x = getopt(me.argc, me.argv, "ndv")) != -1)
  {
    switch (x)
    {
      case 'd':
        me.debug = 1;
        break;
      case 'n':
        no_fork = 1;
        break;
      case 'v':
        printf("%s%s%s by %s\n", VER_STR_BEG, SERIALNUM, VER_STR_END, MYEMAIL);
        printf("Prefix:  %s\n", PREFIX);
        printf("Exiting.\n");
        exit(0);
    }
  }

  if (me.debug == 0)
    fprintf(stderr, "%s%s%s by %s\n", VER_STR_BEG, SERIALNUM, VER_STR_END,
           MYEMAIL);

  InitDefaults();
  ConfLoad();

  VerifyConf();

  if (!(me.logfd = fopen(LOGFILE, "a")))
    fprintf(stderr, "ERROR! Unable to open logfile");

  if ((x = read_db(0)) > 0)
    fprintf(stderr, "ERROR! Unable to read %s", databases[x]);

  if (me.ip.s_addr == INADDR_ANY)
  {
    if (!self_lookup())
      fprintf(stderr, "ERROR! self-lookup failed!");
    else
      fprintf(stderr,"WARNING: no vhost defined in conf, using default of %s\n",
             inet_ntoa(me.ip));
  }

  if (me.debug == 0 && no_fork == 0)
    switch (me.pid = fork())
    {
      case - 1:
        fprintf(stderr, "ERROR! Unable to fork");
        exit(4);
      case 0:
        for (x = 0; x < 12; x++)
          if (x != fileno(me.logfd))
            close(x);
        break;
      default:
        printf("Collaborated IRC Services forked into background: pid = %ld\n",
               (long)me.pid);
        return 0;
    }
  if ((pidf = fopen(PIDFILE, "w")))
  {
    fprintf(pidf, "%ld\n", (long)getpid());
    fclose(pidf);
  }
  else
  {
    slog("ERROR! Unable to open pidfile");
  }

  init_user_hash();

  try_next_hub();
  add_event(1, SERVTIMEOUT / 2, &ping_hub);
  add_event(1, me.conntimeout / 3, &cleanup_dcc);
  add_event(1, 2, &check_midnight);
  /* Starting in 60 seconds from now, check for expiring snuggles every
  ** 5 seconds.
  */
  add_event(time(NULL) + 60, 5, &expire_snuggles);
  io_loop();
  /* NOTREACHED */
  exit(0);
}


void
do_signal(int sig)
{
  switch (sig)
  {
    case SIGHUP:
      tell_chans("SIGHUP received!");
      rehash();
      return ;
    case SIGINT:
    case SIGQUIT:
    case SIGTERM:
      me.crashing = 1;
      write_db(0);
      if (me.servtype == SERV_IRCNN)
        toserv(":%s OPERWALL :services terminating!", me.os_nick, sig);
      else
        toserv(":%s OPERWALL :services terminating!", me.servname, sig);
      toserv(":%s QUIT :services terminating!", me.os_nick);
      fclose(me.logfd);
      close_all_connections();
      exit(1);
#if 0

    case SIGILL:  	/* FALL THROUGH */
    case SIGBUS:
    case SIGSEGV:
    case SIGSYS:
      signal(sig, SIG_DFL);
      if (me.crashing == 1)
        abort();
      me.crashing = 1;
      toserv(":%s OPERWALL :\2FATAL!\2 signal %d recieved!", me.servname, sig);
      kill(getpid(), sig);
#endif

  }
}

/*
 * setup_corefile  
 *
 * inputs       - nothing
 * output       - nothing
 * side effects - setups corefile to system limits.
 * -kre
 *
 * Stolen from Hyb6.2 - Hwy
 */
static void setup_corefile(void)
{
  struct rlimit rlim; /* resource limits */

  /* Set corefilesize to maximum */
  if (!getrlimit(RLIMIT_CORE, &rlim))
  {
    rlim.rlim_cur = rlim.rlim_max;
    setrlimit(RLIMIT_CORE, &rlim);
  }
}

static void check_midnight(void)
{
  static int midnight = 0; /* 1 if the midnight rollover has already run */

  if (!midnight && is_midnight())
  {
    char tmp[BUFSIZE];
    time_t t = time(NULL);

    t = t - 86400;
    /* It is easier to subtract a day here than manipulate variables inside
    ** the localtime struct.
    */

    strftime(tmp, BUFSIZE - 1, LOGFMT, localtime(&t));
    tmp[BUFSIZE - 1] = '\0';

    fclose(me.logfd);
    rename(LOGFILE, tmp);
    me.logfd = fopen(LOGFILE, "a");
    slog("Starting fresh logfile.");
    tell_chans("\2[ROLLOVER]\2 rolling over logfile, its midnight.");
    alldcc("\2[ROLLOVER]\2 rolling over logfile, its midnight.");

    /* It is midnight and the tasks have not yet run. */
    midnight = 1;
  }
  else
  {
    if (midnight && !is_midnight())
    {
      /* It's not midnight anymore, unset that flag */
      midnight = 0;
    }
  }
}

/* Is it the midnight hour */
static int is_midnight(void)
{
  time_t current = time(NULL);
  struct tm *ltm = localtime(&current);

  if (ltm->tm_hour == 0)
    return 1;

  return 0;
}

