/*
*  IRC Services
*
*  Copyright (C) 2003 Services Developers
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of version 1 of the GNU General Public License,
*  as published by the Free Software Foundation.
*
*  This program is distributed in the hope that it will be useful, but
*  WITHOUT ANY WARRANTY; without even the warranty of MERCHANTABILITY
*  or FITNESS FOR A PARTICULAR PURPOSE. The GNU General Public License
*  contains more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
* $Id: users.c,v 1.61 2003/11/05 18:00:35 wcampbel Exp $
*/

#include <assert.h>
#include "irc_types.h"
#include "irc_string.h"
#include "channels.h"
#include "servers.h"
#include "snuggle.h"
#include "defines.h"
#include "client.h"
#include "config.h"
#include "oline.h"
#include "clone.h"
#include "privs.h"
#include "main.h"
#include "send.h"
#include "mem.h"
#include "net.h"
#include "me.h"
#include "operserv.h"
#include "protected.h"

#ifdef XLINE
extern int xline(sock_t *, char *, int, char *);
#endif

user_t *userhash[U_HASH_SIZE];

static user_t *set_user_in_list(user_t *user);

/* Stolen from Sentinel, originally derived from GeoStats */
unsigned int HASH(const unsigned char *name, int size_of_table)
{
  unsigned long h = 0, g;

  while (*name)
  {
    h = (h << 4) + ToLower(*name++);
    if ((g = (h & 0xF0000000)))
      h ^= g >> 24;
    h &= ~g;
  }
  return h % size_of_table;
}

void
init_user_hash(void)
{
  int hc;

  for (hc = 0; hc < U_HASH_SIZE; hc++)
  {
    userhash[hc] = NULL;
  }
}

user_t *
add_user_to_list(char *nick)
{
  user_t *user = leetcalloc(sizeof(user_t), 1);

  user->hash = HASH((unsigned char *) nick, U_HASH_SIZE);
  user->next = userhash[user->hash];
  userhash[user->hash] = user;

  if (user->next)
    user->next->prev = user;
  return user;
}

static user_t *
set_user_in_list(user_t *user)
{
  user->hash = HASH((unsigned char *) user->nick, U_HASH_SIZE);
  user->next = userhash[user->hash];
  user->prev = NULL;
  userhash[user->hash] = user;

  if (user->next)
    user->next->prev = user;
  return user;
}

user_t *
add_user(char *nick, char *username, char *host, char *realname, char *server, int privs)
{
  user_t *user, *sptr = find_client(server);

  if (!sptr)
    return NULL;

  if ((user = find_client(nick)))
    del_user(user);

  user = add_user_to_list(nick);
  user->nick = leetstrdup(nick);
  user->username = leetstrdup(username);
  user->host = leetstrdup(host);
  user->realname = leetstrdup(realname);
  user->server = sptr;

  user->ts = time(NULL);
  user->privs = privs;
  me.users++;

  return user;
}

void
jupe(char *name, char *reason)
{
  user_t *sptr = find_client(name);

  if (sptr && IsServer(sptr))
    squit(sptr->nick, "JUPED");
  add_linked_server(me.servname, name, reason);
  switch (me.servtype)
  {
    case SERV_IRCNN:
      toserv(":%s SERVER %s 2 %lu %lu P09 :%s", me.servname, name, time(NULL), time(NULL) + 1, reason);
      break;
    case SERV_HYBRD:
      toserv(":%s SERVER %s 1 :%s", me.servname, name, reason);
      break;
  }
}

void
add_linked_server(char *hub, char *leaf, char *comment)
{
  user_t *server, *uplink;

  if ((server = find_client(leaf)))
    del_server(server);

  if (!(uplink = find_client(hub)) && !is_me(leaf))
    return ;

  server = add_user_to_list(leaf);
  server->nick = leetstrdup(leaf);
  server->realname = leetstrdup(comment);
  server->server = uplink ? uplink : server;
  server->ts = time(NULL);
  server->privs = SERVER;
  me.servers++;
}

void
del_user_from_list(user_t *user)
{
  if (user->prev != NULL)
  {
    user->prev->next = user->next;
    if (user->next != NULL)
      user->next->prev = user->prev;
  }
  else
  {
    userhash[user->hash] = user->next;
    if (user->next != NULL)
      user->next->prev = NULL;
  }

#if 0
  user->next = NULL;
  user->prev = NULL;
#endif
}

void
del_user(user_t *user)
{
  if (IsServer(user))
  {
    me.servers--;
    del_user_from_list(user);
  }
  else
  {
    if (IsOper(user))
    {
      me.opers--;
    }
    me.users--;
    del_clone_user(user);
    del_user_in_channels(user);
    del_user_from_list(user);
    leetfree(user->host, strlen(user->host) + 1);
    leetfree(user->username, strlen(user->username) + 1);
#ifdef	HAVE_LIBCRYPTO

    if (user->response)
      leetfree(user->response, strlen(user->response) + 1);
#endif

  }
  leetfree(user->nick, strlen(user->nick) + 1);
  leetfree(user->realname, strlen(user->realname) + 1);
  leetfree(user, sizeof(user_t));
}

void
del_all_users(void)
{
  user_t *tmp, *user;
  int i;

  del_all_channels();
  del_all_clones();
  for (i = 0; i < U_HASH_SIZE; i++)
  {
    for (user = userhash[i]; user; )
    {
      leetfree(user->nick, strlen(user->nick) + 1);
      if (user->host)
        leetfree(user->host, strlen(user->host) + 1);
      if (user->username)
        leetfree(user->username, strlen(user->username) + 1);
#ifdef	HAVE_LIBCRYPTO

      if (user->response)
        leetfree(user->response, strlen(user->response) + 1);
#endif

      leetfree(user->realname, strlen(user->realname) + 1);
      tmp = user->next;
      leetfree(user, sizeof(user_t));
      user = tmp;
    }
    userhash[i] = NULL;
  }
  me.admins = me.opers = me.servers = me.gotping = me.eob = me.conn = me.users = 0;
}

user_t *
find_client(char *nick)
{
  unsigned int hc = HASH((unsigned char *) nick, U_HASH_SIZE);
  user_t *user = userhash[hc];

  while (user != NULL)
  {
    /* XXX irccasecmp */
    if (strcasecmp(user->nick, nick) == 0)
      return user;
    user = user->next;
  }
  return NULL;
}

int
change_user_nick(char *old, char *new)
{
  user_t *user = find_client(old);

  if (!user)
    return -1;
  /* Not a leak, restrdup() */
  user->nick = leetrestrdup(user->nick, new);
  if (strcasecmp(old, new) != 0)
    ClearAuth(user);

  /* Move them to the proper place in the hash table */
  del_user_from_list(user);
  set_user_in_list(user);

  if (match_snuggle(user, user->username, user->host, user->realname))
  {
    return 1;
  }
  return 0;
}

int
check_flood(user_t *cptr)
{
  /* Flood checking is disabled */
  if (me.floodrate == 0)
    return 0;

  if (check_protected(cptr))
    return 0;

  if (cptr->ts >= time(NULL))
  {
    if (cptr->flood++ + 1 < me.floodrate)
      return 0;
    if (cptr->flood == me.floodrate)
    {
      cptr->ts = time(NULL) + 10;
      if (!IsOper(cptr))
        serv_notice(cptr->nick, "Flood detected, ignoring you for 10 seconds.");
      else
        tell_chans("\2[FLOOD]\2 %s!%s@%s$%s is flooding services!", cptr->nick, cptr->username, cptr->host, cptr->realname);
    }
    return 1;
  }
  else
  {
    cptr->ts = time(NULL);
    cptr->flood = 0;
    return 0;
  }
}

int
whokill(sock_t *sock, char *nm, char *um, char *hm, char *rm, char *reason)
{
  user_t *user;
  user_t **victim;
  int i, x, hc;

  victim = leetcalloc(me.maxkill + 1, sizeof(user_t));
  i = x = 0;
  for (hc = 0; hc < U_HASH_SIZE; hc++)
  {
    for (user = userhash[hc]; user; user = user->next)
    {
      if (IsPerson(user) && match(nm, user->nick) && match(um, user->username) && match(hm, user->host) && match(rm, user->realname) && user->privs == NOPRIV)
      {
        if (check_protected(user))
        {
          tosock(sock, "User %s!%s@%s$%s is protected, not WHOKILL'ing",
                 user->nick, user->username, user->host, user->realname);
          continue;
        }
        victim[i] = user;
        if (i++ >= me.maxkill)
        {
          alldcc("\2%s\2 attempted to \2WHOKILL\2 over %d people! (%s!%s@%s$%s (%s))",
                 sock->name, me.maxkill, nm, um, hm, rm, reason);
          slog("%s was disconnected for attempting to WHOKILL over %d people!", sock->name, me.maxkill);
          leetfree(victim, ((me.maxkill + 1) * sizeof(user_t)));
          return -5;
        }
      }
    }
  }
  if (!i)
  {
    tosock(sock, "Your WHOKILL request did not match any clients.");
    leetfree(victim, ((me.maxkill + 1) * sizeof(user_t)));
    return 0;
  }
  operwall(OW_CLIENT, "%s requested: WHOKILL %s!%s@%s$%s (%s)", sock->name, nm, um, hm, rm, reason);
  tosock(sock, "Your WHOKILL request matched %d client%s.", i, S(i));
  while (i >= ++x)
  {
    toserv(":%s KILL %s :%s!%s (%s (%d/%d))", me.servname, victim[x - 1]->nick, me.servname, me.os_nick, reason, x, i);
    del_user(victim[x - 1]);
  }
  leetfree(victim, ((me.maxkill + 1) * sizeof(user_t)));
  return i;
}
/* FORBID function here */
#ifdef FORBID
int
forbid(sock_t *sock, char *args)
{}
#endif

int
who(sock_t *sock, char *nm, char *um, char *hm, char *rm)
{
  user_t *user;
  int i = 0, hc;

  for (hc = 0; hc < U_HASH_SIZE; hc++)
  {
    for (user = userhash[hc]; user; user = user->next)
    {
      if (i > 200)
        return i;

      if (IsPerson(user) && match(nm, user->nick) && match(um, user->username) && match(hm, user->host) && match(rm, user->realname))
        tosock(sock, "(%d) %s!%s@%s$%s [%s]", ++i, user->nick, user->username, user->host, user->realname, user->server->nick);
    }
  }
  return i;
}

int
whoc(char *nm, char *um, char *hm, char *rm)
{
  int count = 0;
  user_t *user;
  int hc;

  for (hc = 0; hc < U_HASH_SIZE; hc++)
  {
    for (user = userhash[hc]; user; user = user->next)
    {
      if (IsPerson(user) && match(nm, user->nick) && match(um, user->username) && match(hm, user->host) && match(rm, user->realname))
        count++;
    }
  }
  return count;
}

short
ungline(user_t *cptr, char *um, char *hm)
{
  if (me.servtype == SERV_IRCNN)
    return 1;

  toserv(":%s UNKLINE * %s %s", me.os_nick, um, hm);
  operwall(OW_CLIENT, "\2%s!%s@%s\2 on \2%s\2 issued \2UNGLINE\2 for %s@%s)", cptr->nick, cptr->username, cptr->host, cptr->server->nick, um, hm);
  slog("%s!%s@%s on %s used UNGLINE for %s@%s)", cptr->nick, cptr->username, cptr->host, cptr->server->nick, um, hm);
  return 1;
}

#ifdef XLINE
int
xline(sock_t *sock, char *gecos, int type, char *reason)
/* <sourceoper> XLINE <servermask> <xlinematch> <type> :<reason> */
{
  if (me.servtype == SERV_IRCNN)
  {
    tosock(sock, "XLINE does not work on ircnn.");
    return 1;
  }

  /* me.servtype == SERV_HYB from here on */

  if (type < 0 || type > 2)
  {
    tosock(sock, "XLINE type must be 0, 1, or 2.");
    return 1;
  }

  toserv(":%s XLINE * %s %d :%s", me.os_nick, gecos, type, reason);
  return 0;
}
#endif

int
gline(user_t *cptr, char *um, char *hm, int gtime, char *reason, u_short type)
{
  user_t *user;
  int i = 0, hc;

  if (type == 2)
  {
    if (me.servtype == SERV_IRCNN)
      toserv(":%s KLINE %s@%s %d :AGLINE ID %s -- %s",
             me.servname, um, hm, gtime, me.md5buf, reason);
    else
      toserv(":%s KLINE * %d %s %s :AGLINE ID %s -- %s",
             me.os_nick, gtime * 3600, um, hm, me.md5buf,
             reason);
    return 0;
  }

  if (!cptr)
  {
    if (me.servtype == SERV_IRCNN)
      toserv(":%s KLINE %s@%s %d :%s", me.servname, um, hm, gtime, reason);
    else
      toserv(":%s KLINE * %d %s %s :%s", me.os_nick, gtime * 3600, um, hm, reason);
    return 1;
  }
  i = -1;
  for (hc = 0; hc < U_HASH_SIZE; hc++)
  {
    for (user = userhash[hc]; user; user = user->next)
    {
      if (IsPerson(user) && match(um, user->username) && match(hm, user->host))
      {
        if (i >= me.maxkill)
        {
          slog("%s!%s@%s on %s was KILLED for attempting to GLINE over %d people!", cptr->nick, cptr->username, cptr->host, cptr->server->nick, me.maxkill);
          operwall(OW_CLIENT, "\2%s!%s@%s\2 on \2%s\2 attempted to \2GLINE\2 over %d people! (%s@%s (%s))", cptr->nick, cptr->username, cptr->host, cptr->server->nick, me.maxkill, um, hm, reason);
          irc_kill(cptr->nick, "DO NOT ATTEMPT TO GLINE THAT MANY PEOPLE!@#");
          del_user(cptr);
          return -5;
        }
      }
    }
  }
  if (gtime <= 72)
    operwall(OW_CLIENT, "%s!%s@%s on %s added %sGLINE for %s@%s for %d hour%s (%s)", cptr->nick, cptr->username,
             cptr->host, cptr->server->nick, (type ? "ANON" : ""), um, hm, gtime, S(gtime), reason);
  if (me.servtype == SERV_IRCNN)
    toserv(":%s KLINE %s@%s %d :You have been banned from the network%s%s -- %s [KLINE-ID: %s]", me.servname, um,
           hm, gtime, (type ? "" : " by "), (type ? "" : cptr->nick), reason, me.md5buf);
  else
    toserv(":%s KLINE * %d %s %s :You have been banned from the network%s%s: %s [KLINE-ID: %s]", me.os_nick,
           gtime * 3600, um, hm, (type ? "" : " by "), (type ? "" : cptr->nick), reason, me.md5buf);
  return i;
}

int
del_server(user_t *server)
{
  user_t *tmp, *user;
  int hc;

  assert(0 != server);
  assert(!is_me(server->nick));
  assert(IsServer(server));

  if (!is_me(server->server->nick))
  {
    for (hc = 0; hc < U_HASH_SIZE; hc++)
    {
      for (user = userhash[hc]; user; )
      {
        tmp = user->next;
        if (user->server != server)
        {
          user = tmp;
          continue;
        }
        if (IsServer(user))
          del_server(user);
        else
          del_user(user);
        user = tmp;
      }
    }
  }
  del_user(server);
  return 1;
}


